from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Generic, List, Optional, Type, TypeVar

from schematics import Model

T = TypeVar("T", bound=Model)


@dataclass
class ModelExample(Generic[T]):
    name: str
    value: T
    summary: Optional[str] = None
    description: Optional[str] = None


class ExamplesMixin:
    """
    A mixin that declares that the implementing schema class can supply example objects
    """

    @classmethod
    def get_examples(cls: __class__) -> List[ModelExample[__class__]]:
        """
        Get a collection of examples

        :return: a list of example objects of this class
        """

        raise NotImplementedError()  # Using abc would lead to metaclass conflicts with Schematics


def model_examples_to_openapi_dict(model: Type[ExamplesMixin]) -> Dict[str, Any]:
    """
    Take examples from a model and convert them to OpenAPI examples section.

    :param model: a model to extract
    :return: an OpenAPI examples section
    """

    examples = {}

    for example in model.get_examples():
        examples[example.name] = {"value": example.value.to_primitive()}

        if example.summary is not None:
            examples[example.name]["summary"] = example.summary

        if example.description is not None:
            examples[example.name]["description"] = example.description

    return examples



import inspect
import json
import logging
import re
from copy import deepcopy
from os import path
from typing import Optional, Type

from flask import Blueprint, Flask, Response, jsonify, render_template, request, send_file
from schematics import Model
from schematics.exceptions import DataError
from werkzeug.exceptions import HTTPException

from apistrap.decorators import AcceptsDecorator, RespondsWithDecorator
from apistrap.errors import ApiClientError, ApiServerError, InvalidResponseError, UnexpectedResponseError
from apistrap.extension import Apistrap
from apistrap.schemas import ErrorResponse
from apistrap.types import FileResponse
from apistrap.utils import format_exception, snake_to_camel


class FlaskRespondsWithDecorator(RespondsWithDecorator):
    def _process_response(self, response, is_last_decorator: bool, *args, **kwargs):
        if isinstance(response, Response):
            return response
        if isinstance(response, FileResponse):
            return send_file(
                filename_or_fp=response.filename_or_fp,
                mimetype=self._mimetype or response.mimetype,
                as_attachment=response.as_attachment,
                attachment_filename=response.attachment_filename,
                add_etags=response.add_etags,
                cache_timeout=response.cache_timeout,
                conditional=response.conditional,
                last_modified=response.last_modified,
            )
        if not isinstance(response, self._response_class):
            if is_last_decorator:
                raise UnexpectedResponseError(type(response))
            return response  # Let's hope the next RespondsWithDecorator takes care of the response

        try:
            response.validate()
        except DataError as ex:
            raise InvalidResponseError(ex.errors) from ex

        response = jsonify(response.to_primitive())
        response.status_code = self._code
        return response


class FlaskAcceptsDecorator(AcceptsDecorator):
    def _get_request_content_type(self, *args, **kwargs):
        return request.content_type

    def _get_request_json(self, *args, **kwargs):
        try:
            if request.json is None or isinstance(request.json, str):
                raise ApiClientError("The request body must be a JSON object")
        except json.decoder.JSONDecodeError as ex:
            raise ApiClientError("The request body must be a JSON object") from ex

        return request.json


class FlaskApistrap(Apistrap):
    PARAMETER_TYPE_MAP = {int: "integer", str: "string"}

    def __init__(self):
        super().__init__()
        self._app: Flask = None
        self._specs_extracted = False

    def init_app(self, app: Flask):
        """
        Bind the extension to a Flask instance.

        :param app: the Flask instance
        """
        self._app = app
        blueprint = Blueprint("apistrap", __name__, template_folder=path.join(path.dirname(__file__), "templates"))

        if self.spec_url is not None:
            blueprint.route(self.spec_url, methods=["GET"])(self._get_spec)

        if self.spec_url is not None and self.ui_url is not None:
            blueprint.route(self.ui_url)(self._get_ui)
            blueprint.route(self.ui_url + "/")(self._get_ui)

        app.register_blueprint(blueprint)

        if self.use_default_error_handlers:
            app.register_error_handler(HTTPException, self.http_error_handler)
            app.register_error_handler(ApiClientError, self.error_handler)
            app.register_error_handler(ApiServerError, self.internal_error_handler)
            app.register_error_handler(Exception, self.internal_error_handler)

    def _is_bound(self) -> bool:
        return self._app is not None

    def _get_spec(self):
        """
        Serves the OpenAPI specification
        """
        self._extract_specs()
        return jsonify(self.to_openapi_dict())

    _get_spec.apistrap_ignore = True

    def _get_ui(self):
        """
        Serves Swagger UI
        """
        return render_template("apidocs.html", apistrap=self)

    _get_ui.apistrap_ignore = True

    def _extract_specs(self):
        """
        Extract specification data from the Flask app and save it to the underlying Apispec object
        """
        if self._specs_extracted:
            return

        for rule in self._app.url_map.iter_rules():
            # Skip Flask's internal static endpoint
            if rule.endpoint == "static":
                continue

            handler = self._app.view_functions[rule.endpoint]

            # Skip ignored endpoints
            if getattr(handler, "apistrap_ignore", False):
                continue

            url = str(rule)
            for arg in re.findall("(<([^<>]*:)?([^<>]*)>)", url):
                url = url.replace(arg[0], "{%s}" % arg[2])

            for method in rule.methods:
                if method.lower() not in ["get", "post", "put", "delete", "patch"]:
                    continue

                self.spec.path(url, {method.lower(): self._extract_operation_specs(handler)})

        self._specs_extracted = True

    def _extract_operation_specs(self, handler):
        """
        Extract operation specification data from a Flask view handler

        :param handler: the Flask handler to extract
        :return: a dictionary containing the specification data
        """

        specs_dict = deepcopy(getattr(handler, "specs_dict", {"parameters": [], "responses": {}}))
        specs_dict["summary"] = handler.__doc__.strip() if handler.__doc__ else ""

        signature = inspect.signature(handler)
        ignored = getattr(handler, "_ignored_params", [])

        for arg in signature.parameters.values():
            if arg.name not in ignored:
                param_data = {
                    "in": "path",
                    "name": arg.name,
                    "required": True,
                    "schema": {"type": self.PARAMETER_TYPE_MAP.get(arg.annotation, "string")},
                }

                specs_dict["parameters"].append(param_data)

        specs_dict["operationId"] = snake_to_camel(handler.__name__)

        return specs_dict

    def http_error_handler(self, exception: HTTPException):
        """
        A handler for Flask HTTPException objects
        :param exception: the exception raised due to a client error
        :return: a response object
        """

        logging.exception(exception)
        return jsonify(ErrorResponse(dict(message=exception.description)).to_primitive()), exception.code

    def error_handler(self, exception):
        """
        The default handler for API client errors
        :param exception: the exception raised due to a client error
        :return: a response object
        """

        if self._app.debug:
            logging.exception(exception)
            error_response = ErrorResponse(dict(message=str(exception), debug_data=format_exception(exception)))
        else:
            error_response = ErrorResponse(dict(message=str(exception)))

        return jsonify(error_response.to_primitive()), 400

    def internal_error_handler(self, exception):
        """
        The default handler for API server errors. It will hide the details when not in debug mode.
        :param exception: the exception raised due to a server error
        :return: a response object
        """

        logging.exception(exception)

        if self._app.debug:
            error_response = ErrorResponse(dict(message=str(exception), debug_data=format_exception(exception)))
        else:
            error_response = ErrorResponse(dict(message="Internal server error"))

        return jsonify(error_response.to_primitive()), 500

    def responds_with(
        self,
        response_class: Type[Model],
        *,
        code: int = 200,
        description: Optional[str] = None,
        mimetype: Optional[str] = None
    ):
        return FlaskRespondsWithDecorator(self, response_class, code=code, description=description, mimetype=mimetype)

    def accepts(self, request_class: Type[Model]):
        return FlaskAcceptsDecorator(self, request_class)



from typing import Dict, List


class ApistrapExtensionError(Exception):
    """
    Raised on incorrect usage of the Apistrap extension
    """


class ApiClientError(Exception):
    """
    An exception raised when the user of the API makes an invalid request
    """


class InvalidFieldsError(ApiClientError):
    """
    An exception raised when the user of the API sends a malformed request object
    """

    def __init__(self, errors: Dict[str, List[str]]):
        super().__init__(f"Invalid input: `{str(errors)}`")
        self.errors = errors


class ApiServerError(Exception):
    """
    An exception raised when the server encounters an error
    """


class UnexpectedResponseError(ApiServerError):
    """
    An exception raised when a view function returns a response of an unexpected type
    """

    def __init__(self, response_class: type):
        super().__init__(f"Unexpected response class: `{response_class.__name__}`")


class InvalidResponseError(ApiServerError):
    """
    An exception raised when a view function returns a malformed response
    """

    def __init__(self, errors: Dict[str, List[str]]):
        super().__init__(f"Invalid input: `{str(errors)}`")
        self.errors = errors



import json
import logging
import mimetypes
import re
from copy import deepcopy
from itertools import chain
from os import path
from pathlib import Path
from typing import Any, Callable, Coroutine, Optional, Tuple, Type

import jinja2
from aiohttp import web
from aiohttp.web_exceptions import HTTPError
from aiohttp.web_request import BaseRequest, Request
from aiohttp.web_response import Response
from aiohttp.web_urldispatcher import AbstractRoute, DynamicResource, PlainResource
from schematics import Model
from schematics.exceptions import DataError

from apistrap.decorators import AcceptsDecorator, RespondsWithDecorator
from apistrap.errors import ApiClientError, InvalidResponseError, UnexpectedResponseError
from apistrap.extension import Apistrap
from apistrap.schemas import ErrorResponse
from apistrap.types import FileResponse
from apistrap.utils import format_exception, snake_to_camel


class AioHTTPRespondsWithDecorator(RespondsWithDecorator):
    async def _process_response(self, response, is_last_decorator: bool, *args, **kwargs):
        if isinstance(response, Response) or isinstance(response, web.StreamResponse):
            return response
        if isinstance(response, FileResponse):
            # TODO consider implementing add_etags, cache_timeout and conditional
            headers = {}

            if self._mimetype:
                headers["Content-Type"] = self._mimetype
            elif response.mimetype:
                headers["Content-Type"] = response.mimetype
            elif response.attachment_filename:
                headers["Content-Type"] = mimetypes.guess_type(response.attachment_filename)[0]

            if response.last_modified is not None:
                headers["Last-Modified"] = str(response.last_modified)

            if response.as_attachment:
                if response.attachment_filename is None:
                    raise TypeError("Missing attachment filename")

                headers["Content-Disposition"] = f"attachment,filename={response.attachment_filename}"

            if isinstance(response.filename_or_fp, str) or isinstance(response.filename_or_fp, Path):
                return web.FileResponse(response.filename_or_fp, headers=headers)
            else:
                stream = web.StreamResponse(headers=headers)
                request = next(filter(lambda a: isinstance(a, BaseRequest), chain(args, kwargs.values())), None)

                if request is None:
                    raise TypeError("No request passed to view function")

                await stream.prepare(request)

                while True:
                    chunk = response.filename_or_fp.read(16536)
                    if not chunk:
                        await stream.write_eof()
                        break

                    await stream.write(chunk)

                return stream

        if not isinstance(response, self._response_class):
            if is_last_decorator:
                raise UnexpectedResponseError(type(response))
            return response  # Let's hope the next RespondsWithDecorator takes care of the response

        try:
            response.validate()
        except DataError as e:
            raise InvalidResponseError(e.errors) from e

        return web.Response(
            text=json.dumps(response.to_primitive()), content_type="application/json", status=self._code
        )


class AioHTTPAcceptsDecorator(AcceptsDecorator):
    def _get_request_content_type(self, request: BaseRequest, *args, **kwargs):
        return request.content_type

    async def _get_request_json(self, request: BaseRequest, *args, **kwargs):
        try:
            data = await request.json()
        except json.decoder.JSONDecodeError as ex:
            raise ApiClientError("The request body must be a JSON object") from ex

        if isinstance(data, str):
            raise ApiClientError("The request body must be a JSON object")

        return data


ErrorHandler = Callable[[Exception], Tuple[ErrorResponse, int]]


@web.middleware
class ErrorHandlerMiddleware:
    """
    A configurable handler for exceptions raised when processing HTTP requests.
    """

    def __init__(self, apistrap: "AioHTTPApistrap"):
        """
        :param apistrap: The apistrap extension object
        """
        self._apistrap = apistrap
        self._handlers = []
        self._default_handlers = [
            (HTTPError, self._handle_http_error),
            (ApiClientError, self._handle_client_error),
            (Exception, self._handle_server_error),
        ]

    def add_handler(self, exception_type: Type[Exception], handler: ErrorHandler) -> None:
        """
        Add a new error handler.

        :param exception_type: Only instances of this exception type will be handled by the handler
        :param handler: A function that handles the exception and returns a response
        """
        self._handlers.append((exception_type, handler))

    def handle_error(self, exception: Exception) -> Tuple[ErrorResponse, int]:
        """
        Find a handler for an exception and invoke it, then return an error response.

        :param exception: the exception to be handled
        :return: an ErrorResponse instance
        """
        for handled_type, handler in chain(
            self._handlers, self._default_handlers if self._apistrap.use_default_error_handlers else []
        ):
            if isinstance(exception, handled_type):
                return handler(exception)

        raise ValueError(f"Unexpected exception type `{type(exception).__name__}`") from exception

    def _handle_server_error(self, exception):
        """
        Default handler for server errors (500-599).

        :param exception: the exception to be handled
        :return: an ErrorResponse instance
        """
        logging.exception(exception)
        if self._apistrap.app.debug:
            return ErrorResponse(dict(message=str(exception), debug_data=format_exception(exception))), 500
        else:
            return ErrorResponse(dict(message="Internal server error")), 500

    def _handle_client_error(self, exception):
        """
        Default handler for client errors (400-499).

        :param exception: the exception to be handled
        :return: an ErrorResponse instance
        """
        if self._apistrap.app.debug:
            logging.exception(exception)
            return ErrorResponse(dict(message=str(exception), debug_data=format_exception(exception))), 400
        else:
            return ErrorResponse(dict(message=str(exception))), 400

    def _handle_http_error(self, exception):
        """
        Default handler for http errors (e.g. 404).

        :param exception: the exception to be handled
        :return: an ErrorResponse instance
        """
        logging.exception(exception)
        return ErrorResponse(dict(message=exception.text)), exception.status_code

    async def __call__(
        self, request: BaseRequest, handler: Callable[[BaseRequest], Coroutine[Any, Any, Response]]
    ) -> web.Response:
        """
        Invoke the middleware.

        :param request: the API request (will be handled without any modifications)
        :param handler: the handler function (will be invoked without any modifications)
        :return: either the response of the handler or an error message response
        """
        try:
            return await handler(request)
        except Exception as ex:
            error_response, code = self.handle_error(ex)

            return web.Response(
                text=json.dumps(error_response.to_primitive()), content_type="application/json", status=code
            )


class AioHTTPApistrap(Apistrap):
    def __init__(self):
        super().__init__()
        self.app: web.Application = None
        self.error_middleware = ErrorHandlerMiddleware(self)
        self._specs_extracted = False
        self._jinja_env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(path.join(path.dirname(__file__), "templates"))
        )

    def init_app(self, app: web.Application) -> None:
        """
        Bind the extension to an AioHTTP instance.

        :param app: the AioHTTP instance
        """

        self.app = app
        app.middlewares.append(self.error_middleware)

        if self.spec_url is not None:
            app.router.add_route("get", self.spec_url, self._get_spec)

            if self.ui_url is not None:
                for ui_url in (self.ui_url, self.ui_url + "/"):
                    app.router.add_route("get", ui_url, self._get_ui)

    def _get_spec(self, request: Request):
        """
        Serves the OpenAPI specification
        """

        self._extract_specs()
        return web.Response(text=json.dumps(self.to_openapi_dict()), content_type="application/json", status=200)

    _get_spec.apistrap_ignore = True

    def _get_ui(self, request: Request):
        """
        Serves Swagger UI
        """

        return web.Response(
            text=self._jinja_env.get_template("apidocs.html").render(apistrap=self),
            content_type="text/html",
            status=200,
        )

    _get_ui.apistrap_ignore = True

    def _extract_specs(self) -> None:
        """
        Extract the specification data from the bound AioHTTP app. If the data was already extracted, do not do
        anything.
        """

        if self._specs_extracted:
            return

        route: AbstractRoute
        for route in self.app.router.routes():
            if getattr(route.handler, "apistrap_ignore", False):
                continue

            if route.method.lower() not in ["get", "post", "put", "delete", "patch"]:
                continue

            url = ""

            if isinstance(route.resource, PlainResource):
                url = route.resource.get_info()["path"]
            elif isinstance(route.resource, DynamicResource):
                url = route.resource.get_info()["formatter"]

            self.spec.path(url, {route.method.lower(): self._extract_operation_spec(route)})

        self._specs_extracted = True

    def _extract_operation_spec(self, route: AbstractRoute) -> dict:
        """
        Extract specification data for a single operation.

        :param route: the route for the operation
        :return: a dict with specification data
        """

        specs_dict = deepcopy(getattr(route.handler, "specs_dict", {"parameters": [], "responses": {}}))
        specs_dict["summary"] = route.handler.__doc__.strip() if route.handler.__doc__ else ""

        if isinstance(route.resource, DynamicResource):
            parameters = re.findall(r"{([a-zA-Z0-9]+[^}]*)}", route.resource.get_info()["formatter"])

            for arg in parameters:
                param_data = {
                    "in": "path",
                    "name": arg,
                    "required": True,
                    "schema": {"type": "string"},  # TODO support other types too
                }

                specs_dict["parameters"].append(param_data)

        specs_dict["operationId"] = snake_to_camel(route.handler.__name__)

        return specs_dict

    def _is_bound(self) -> bool:
        return self.app is not None

    def responds_with(
        self,
        response_class: Type[Model],
        *,
        code: int = 200,
        description: Optional[str] = None,
        mimetype: Optional[str] = None,
    ):
        """
        A decorator that fills in response schemas in the OpenAPI specification. It also converts Schematics models
        returned by view functions to JSON and validates them.
        """
        return AioHTTPRespondsWithDecorator(self, response_class, code=code, description=description, mimetype=mimetype)

    def accepts(self, request_class: Type[Model]):
        """
        A decorator that validates request bodies against a schema and passes it as an argument to the view function.
        The destination argument must be annotated with the request type.
        """
        return AioHTTPAcceptsDecorator(self, request_class)



from inspect import getmro
from typing import Any, Dict, Generator, Optional, Type

from schematics import Model
from schematics.types.base import (BaseType, BooleanType, DecimalType, FloatType, IntType, LongType, NumberType,
                                   StringType)
from schematics.types.compound import DictType, ListType, ModelType


def _get_serialized_name(field: BaseType) -> str:
    """
    Get the name under which the model field is saved in the serialized object

    :param field: the model field
    :return: a field name
    """
    return getattr(field, "serialized_name", None) or field.name


def _model_fields_to_schema_object_properties(model: Type[Model]) -> Dict[str, Any]:
    """
    Convert all fields of a model to OpenAPI 3 SchemaObject Properties objects

    :param model: the model to be converted
    :return: a dictionary with field names as keys and SchemaObjects as values
    """
    properties = {}

    for field in model._fields.values():
        serialized_name = _get_serialized_name(field)
        schema_object = _field_to_schema_object(field)

        if schema_object is not None:
            properties[serialized_name] = schema_object

    return properties


def _field_to_schema_object(field: BaseType) -> Optional[Dict[str, Any]]:
    """
    Convert a field definition to OpenAPI 3 schema.

    :param field: the field to be converted
    :return: a schema
    """

    if isinstance(field, ModelType):
        return _model_field_to_schema_object(field)
    elif isinstance(field, ListType):
        if isinstance(field.field, ModelType):
            return _model_array_to_schema_object(field)
        elif isinstance(field.field, BaseType):
            return _primitive_array_to_schema_object(field)
    elif isinstance(field, DictType):
        if isinstance(field.field, ModelType):
            return _model_dict_to_schema_object(field)
        elif isinstance(field.field, BaseType):
            return _primitive_dict_to_schema_object(field)
    elif isinstance(field, StringType):
        return _string_field_to_schema_object(field)
    elif isinstance(field, BaseType):
        return _primitive_field_to_schema_object(field)

    return None


SCHEMATICS_TYPE_TO_JSON_TYPE = {
    NumberType: "number",
    IntType: "integer",
    LongType: "integer",
    FloatType: "number",
    DecimalType: "number",
    BooleanType: "boolean",
}

SCHEMATICS_OPTIONS_TO_JSON_SCHEMA = {
    "max_length": "maxLength",
    "min_length": "minLength",
    "regex": "pattern",
    "min_value": "minimum",
    "max_value": "maximum",
    "min_size": "minItems",
    "max_size": "maxItems",
}


def _get_field_type(field: BaseType):
    for cls in getmro(field.__class__):
        if cls in SCHEMATICS_TYPE_TO_JSON_TYPE.keys():
            return SCHEMATICS_TYPE_TO_JSON_TYPE[cls]

    return "string"


def _extract_model_description(field: BaseType) -> Dict[str, Any]:
    """
    Load title and description from Schematics metadata.

    :param field: field to extract metadata from
    :return a dict with optional title and description keys
    """

    metadata = field.metadata
    data = {}

    if "label" in metadata:
        data["title"] = metadata["label"]

    if "description" in metadata:
        data["description"] = metadata["description"]

    return data


def _primitive_field_to_schema_object(field: BaseType) -> Dict[str, str]:
    schema = {"type": _get_field_type(field)}

    schema.update(_extract_model_description(field))

    for schematics_attr, json_schema_attr in SCHEMATICS_OPTIONS_TO_JSON_SCHEMA.items():
        if hasattr(field, schematics_attr):
            option_value = getattr(field, schematics_attr)
            if option_value is not None:
                schema[json_schema_attr] = option_value

    return schema


def _string_field_to_schema_object(field: StringType) -> Dict[str, str]:
    schema = _primitive_field_to_schema_object(field)

    if field.choices is not None:
        schema["enum"] = field.choices

    return schema


def _get_required_fields(model: Type[Model]) -> Generator[str, None, None]:
    """
    Extract required field names from a Schematics model

    :param model:
    :return: a generator of field names
    """

    for field in model._fields.values():
        if getattr(field, "required", False):
            yield _get_serialized_name(field)


def _model_array_to_schema_object(field: ModelType) -> Dict[str, Any]:
    """
    Get a SchemaObject for a list of given models

    :param field: the field field to be converted
    :return: a SchemaObject
    """

    model = field.field.model_class

    schema = {"type": "array", "title": f"List of {model.__name__}", "items": schematics_model_to_schema_object(model)}

    schema.update(_extract_model_description(field))

    return schema


def _model_dict_to_schema_object(field: DictType) -> Dict[str, Any]:
    """
    Get a SchemaObject for a dictionary of given models

    :param field: the field to be converted
    :return: a SchemaObject
    """

    model = field.field.model_class

    schema = {
        "type": "object",
        "title": f"Dictionary of {model.__name__}",
        "additionalProperties": schematics_model_to_schema_object(model),
    }

    schema.update(_extract_model_description(field))

    return schema


def _primitive_array_to_schema_object(field: ListType) -> Dict[str, Any]:
    """
    Get a SchemaObject for a list of primitive types

    :param field: the field that determines the value type
    :return: a SchemaObject
    """

    schema = {
        "type": "array",
        "title": f"List of {field.field.__class__.__name__}",
        "items": _field_to_schema_object(field.field),
    }

    schema.update(_extract_model_description(field))

    return schema


def _primitive_dict_to_schema_object(field: DictType) -> Dict[str, Any]:
    """
    Get a SchemaObject for a dictionary of primitive types

    :param field: the field that determines the value type
    :return: a SchemaObject
    """

    schema = {
        "type": "object",
        "title": f"Dictionary of {field.field.__class__.__name__}",
        "additionalProperties": _field_to_schema_object(field.field),
    }

    schema.update(_extract_model_description(field))

    return schema


def _model_field_to_schema_object(field: ModelType) -> Dict[str, Any]:
    """
    Get a SchemaObject for a model field.

    :param field: the field that determines the value type
    :return: a SchemaObject
    """

    schema = schematics_model_to_schema_object(field.model_class)
    schema.update(_extract_model_description(field))

    return schema


def schematics_model_to_schema_object(model: Type[Model]) -> Dict[str, Any]:
    """
    Convert a Schematics model to an OpenAPI 3 SchemaObject

    :param model: the model to be converted
    :return: a SchemaObject
    """
    schema = {
        "type": "object",
        "title": model.__name__,
        "properties": (_model_fields_to_schema_object_properties(model)),
    }

    required = list(_get_required_fields(model))

    if required:
        schema["required"] = required

    return schema



import traceback
from typing import Any, Mapping

from more_itertools import flatten


def format_exception(exception: Exception) -> Mapping[str, Any]:
    """
    Format an exception into a dict containing exception information such as class name, message and traceback.

    Example:

        {
            'exception_type': 'RuntimeError'
            'exception_message': 'Cannot connect to database.',
            'traceback': ['  File "spam.py", line 3, in <module>',
                          '   spam.eggs()\n',
                          ...]
        }
    """

    traceback_summary = traceback.StackSummary.extract(traceback.walk_tb(exception.__traceback__))

    return {
        "exception_type": type(exception).__name__,
        "exception_message": str(exception),
        "traceback": list(flatten([str(s).rstrip("\n").split("\n") for s in traceback_summary.format()])),
    }


def snake_to_camel(value):
    """
    Convert a string from snake_case to camelCase
    """
    result = "".join(x.capitalize() or "_" for x in value.split("_"))
    return result[0].lower() + result[1:]



from __future__ import annotations

import abc
import inspect
from functools import wraps
from typing import TYPE_CHECKING, Callable, Optional, Sequence, Type

from schematics import Model
from schematics.exceptions import DataError

from apistrap.errors import ApiClientError, InvalidFieldsError
from apistrap.examples import ExamplesMixin, model_examples_to_openapi_dict
from apistrap.schematics_converters import schematics_model_to_schema_object
from apistrap.types import FileResponse

if TYPE_CHECKING:
    from apistrap.extension import Apistrap


def _ensure_specs_dict(func: Callable):
    if not hasattr(func, "specs_dict"):
        func.specs_dict = {"parameters": [], "responses": {}}


def _add_ignored_param(func: Callable, arg: str):
    if not hasattr(func, "_ignored_params"):
        setattr(func, "_ignored_params", [])

    func._ignored_params.append(arg)


def _get_wrapped_function(func: Callable):
    """
    Get the actual function from a decorated function. This could end up in a loop on horribly mangled functions.
    """

    wrapped = getattr(func, "__wrapped__", None)

    if wrapped is None:
        return func

    return _get_wrapped_function(wrapped)


class IgnoreParamsDecorator:
    """
    A decorator that marks specified function parameters as ignored for the purposes of generating a specification
    """

    def __init__(self, ignored_params: Sequence[str]):
        self._ignored_params = ignored_params

    def __call__(self, wrapped_func):
        for param in self._ignored_params:
            _add_ignored_param(wrapped_func, param)
        return wrapped_func


class RespondsWithDecorator:
    """
    A decorator that fills in response schemas in the Swagger specification. It also converts Schematics models returned
    by view functions to JSON and validates them.
    """

    outermost_decorators = {}
    """
    Maps functions to the outermost RespondsWithDecorator so that we can perform checks for unknown response classes 
    when we get to the last decorator
    """

    def __init__(
        self,
        apistrap: Apistrap,
        response_class: Type[Model],
        *,
        code: int = 200,
        description: Optional[str] = None,
        mimetype: Optional[str] = None,
    ):
        self._response_class = response_class
        self._code = code
        self._apistrap = apistrap
        self._description = description or self._response_class.__name__
        self._mimetype = mimetype

    def __call__(self, wrapped_func: Callable):
        _ensure_specs_dict(wrapped_func)

        if self._response_class == FileResponse:
            wrapped_func.specs_dict["responses"][str(self._code)] = {
                "description": self._description or self._response_class.__name__,
                "content": {
                    self._mimetype or "application/octet-stream": {"schema": {"type": "string", "format": "binary"}}
                },
            }
        else:
            wrapped_func.specs_dict["responses"][str(self._code)] = {
                "description": self._description or self._response_class.__name__,
                "content": {
                    "application/json": {
                        "schema": {
                            "$ref": self._apistrap.add_schema_definition(
                                self._response_class.__name__, self._get_schema_object()
                            )
                        }
                    }
                },
            }

            if issubclass(self._response_class, ExamplesMixin):
                # fmt: off
                wrapped_func.specs_dict["responses"][str(self._code)]["content"]["application/json"]["examples"] = \
                    model_examples_to_openapi_dict(self._response_class)
                # fmt: on

        innermost_func = _get_wrapped_function(wrapped_func)
        self.outermost_decorators[innermost_func] = self

        if inspect.iscoroutinefunction(wrapped_func):

            @wraps(wrapped_func)
            async def wrapper(*args, **kwargs):
                response = await wrapped_func(*args, **kwargs)
                is_last_decorator = self.outermost_decorators[innermost_func] == self
                return await self._process_response(response, is_last_decorator, *args, **kwargs)

        else:

            @wraps(wrapped_func)
            def wrapper(*args, **kwargs):
                response = wrapped_func(*args, **kwargs)
                is_last_decorator = self.outermost_decorators[innermost_func] == self
                return self._process_response(response, is_last_decorator)

        return wrapper

    def _get_schema_object(self):
        return schematics_model_to_schema_object(self._response_class)

    def _process_response(self, response, is_last_decorator: bool, *args, **kwargs):
        """
        Process a response received from an endpoint handler (i.e. send it)
        :param response: the response to be processed
        :param is_last_decorator: True if the current decorator is the outermost one
        """


class AcceptsDecorator(metaclass=abc.ABCMeta):
    """
    A decorator that validates request bodies against a schema and passes it as an argument to the view function.
    The destination argument must be annotated with the request type.
    """

    def __init__(self, apistrap: Apistrap, request_class: Type[Model]):
        self._apistrap = apistrap
        self._request_class = request_class

    def _find_parameter_by_request_class(self, signature: inspect.Signature) -> Optional[inspect.Parameter]:
        for parameter in signature.parameters.values():
            if isinstance(parameter.annotation, str):
                if parameter.annotation == self._request_class.__qualname__:
                    return parameter
            elif isinstance(parameter.annotation, type):
                if issubclass(self._request_class, parameter.annotation):
                    return parameter
        return None

    def __call__(self, wrapped_func: Callable):
        _ensure_specs_dict(wrapped_func)

        # TODO parse title from param in docblock
        wrapped_func.specs_dict["requestBody"] = {
            "content": {
                "application/json": {
                    "schema": {
                        "$ref": self._apistrap.add_request_definition(
                            self._request_class.__name__, schematics_model_to_schema_object(self._request_class)
                        )
                    }
                }
            },
            "required": True,
        }

        if issubclass(self._request_class, ExamplesMixin):
            # fmt: off
            wrapped_func.specs_dict["requestBody"]["content"]["application/json"]["examples"] = \
                model_examples_to_openapi_dict(self._request_class)
            # fmt: on

        wrapped_func.specs_dict["x-codegen-request-body-name"] = "body"

        signature = inspect.signature(wrapped_func)
        request_arg = self._find_parameter_by_request_class(signature)

        if request_arg is None:
            raise TypeError(f"no argument of type `{self._request_class}` found")

        if inspect.iscoroutinefunction(wrapped_func):

            @wraps(wrapped_func)
            async def wrapper(*args, **kwargs):
                self._check_request_type(*args, **kwargs)
                body = await self._get_request_json(*args, **kwargs)
                kwargs = self._process_request_kwargs(body, signature, request_arg, *args, **kwargs)
                return await wrapped_func(*args, **kwargs)

        else:

            @wraps(wrapped_func)
            def wrapper(*args, **kwargs):
                self._check_request_type(*args, **kwargs)
                body = self._get_request_json()
                kwargs = self._process_request_kwargs(body, signature, request_arg, *args, **kwargs)
                return wrapped_func(*args, **kwargs)

        _add_ignored_param(wrapper, request_arg.name)
        return wrapper

    def _check_request_type(self, *args, **kwargs):
        if self._get_request_content_type(*args, **kwargs) != "application/json":
            raise ApiClientError("Unsupported media type, JSON is expected")

    def _process_request_kwargs(self, body, signature, request_arg, *args, **kwargs):
        bound_args = signature.bind_partial(*args, **kwargs)
        if request_arg.name not in bound_args.arguments:
            request_object = self._request_class.__new__(self._request_class)

            try:
                request_object.__init__(body, validate=True, partial=False, strict=True)
            except DataError as e:
                raise InvalidFieldsError(e.errors) from e

            new_kwargs = {request_arg.name: request_object}
            new_kwargs.update(**kwargs)
            return new_kwargs

    @abc.abstractmethod
    def _get_request_content_type(self, *args, **kwargs) -> str:
        """
        Get the value of the Content-Type header of current request
        """

    @abc.abstractmethod
    def _get_request_json(self, *args, **kwargs):
        """
        Get the JSON content of the request
        """


class IgnoreDecorator:
    """
    A decorator that marks an endpoint as ignored so that the extension won't include it in the specification.
    """

    def __call__(self, wrapped_func: Callable):
        wrapped_func.apistrap_ignore = True
        return wrapped_func


class TagsDecorator:
    """
    A decorator that adds tags to the OpenAPI specification of the decorated view function.
    """

    def __init__(self, tags: Sequence[str]):
        self._tags = tags

    def __call__(self, wrapped_func: Callable):
        _ensure_specs_dict(wrapped_func)
        wrapped_func.specs_dict.setdefault("tags", [])
        wrapped_func.specs_dict["tags"].extend(self._tags)
        return wrapped_func


class SecurityDecorator:
    """
    A decorator that enforces user authentication and authorization.
    """

    def __init__(self, extension: Apistrap, scopes: Sequence[str]):
        self._extension = extension
        self._scopes = scopes

    def __call__(self, wrapped_func: Callable):
        _ensure_specs_dict(wrapped_func)
        wrapped_func.specs_dict.setdefault("security", [])

        for scheme in self._extension.security_schemes:
            wrapped_func.specs_dict["security"].append({scheme.name: [*map(str, self._scopes)]})

            wrapped_func = scheme.enforcer(self._scopes)(wrapped_func)

        return wrapped_func
